<div class="col-md-{% if 'full_width' in metric %}12{% else %}6{% endif %} mb-5">
    <a name="{% if feature%}{{ feature.name }}-{%endif%}{{ metric.name }}"></a>
    <div class="card shadow-sm">
        <div class="card-body" style="text-align: center">
            <h4 class="card-title">{{metric.name | fmt_metric}}</h4>
            {% if metric.description|length %}
                <p class="card-text">{{metric.description}}</p>
            {% endif %}
        </div>
        {% if 'table' in metric.plot %}
            {{ metric.plot }}
        {% else %}
        <img class="card-img-top" src="data:image/png;base64,{{metric.plot}}" alt="" />
        {% endif %}
    </div>
</div>