body {
    padding-top: 56px;
}

header {
  padding: 156px 0 100px;
}

section {
  padding: 150px 0;
}

.card-footer {
  text-align: center;
}

section {
  padding: 70px 0
}

a.nav-link {
  cursor: pointer;
  font-weight: 300;
}

.navbar-brand {
  color: #fff !important;
  font-weight: 700
}

.card-body {
  padding-bottom: 0 !important
}

p.card-text {
  font-size: 11px;
  font-weight: 300;
}

div.section-description {
  margin: 20px 0;
  font-size: 14px;
  font-weight: 300;
}

section h2 {
  font-weight: 700
}

.feature-select button {
  font-weight: 300
}

.card-image-top {
  padding-bottom: 7px;
}

/* overview tables */
table.overview{
    margin: 25px;
}
table.overview tbody td.metric{
    white-space: nowrap;
}
table.overview tbody td.cell{
   border: 1px solid #333333;
   text-align: center;
}
table.overview td.cell-green{
    background: green;
}
table.overview td.cell-yellow{
    background: rgba(255, 200, 0, 1.0);
}
table.overview td.cell-red{
    background: rgba(255, 0, 0, 1.0);
}
table.overview tfoot td {
    padding-top: 5px;
    text-align: center;
}
table.overview tfoot td span{
    -ms-writing-mode: tb-rl;
    -webkit-writing-mode: vertical-rl;
    writing-mode: vertical-rl;
    transform: rotate(180deg);
    white-space: nowrap;

    font-size: 14px;
    font-weight: 300;
}

/* overview progress bars */
.green, .yellow, .red, .white{
    float:left;
    padding-top:10px;
    padding-bottom:10px;
    color: white;
    text-align:center;
}
.green{
    background-color: green;
}
.yellow{
    background-color: rgba(255, 200, 0, 1.0);
}
.red{
    background-color: rgba(255, 0, 0, 1.0);
}
