<!-- Footer -->
<footer class="py-5 bg-dark">
    <div class="container">
        <p class="m-0 text-center text-white">Copyright &copy; ING WBAA 2022</p>
    </div>
    <!-- /.container -->
</footer>

<!-- Bootstrap core JavaScript -->
<!-- Plugin JavaScript -->
<!-- Custom JavaScript for this theme -->
<script>
    {% include 'assets/js/jquery.min.js' %}
    {% include 'assets/js/bootstrap.bundle.min.js' %}
    {% include 'assets/js/jquery.easing.min.js' %}
    {% include 'assets/js/scrolling-nav.js' %}
    {% include 'assets/js/custom-script.js' %}
</script>