<!DOCTYPE html>
<html lang="en">
<head>

  <meta charset="utf-8">
  <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">
  <meta name="generator" content="{{ generator }}" />
  <meta name="description" content="Automatically generated population shift report" />
  <meta name="url" content="https://github.com/ing-bank/popmon" />

  <title>POPMON Report</title>

  <!-- Bootstrap core CSS -->
  <!-- Custom styles for this template -->
  <style>
    {% include 'assets/css/bootstrap.min.css' %}
    {% include 'assets/css/custom-style.css' %}
  </style>

</head>
<body id="page-top">

    {% include 'modal-popup.html' %}
    {% include 'header.html' %}
    {{ sections }}
    {% include 'footer.html' %}

</body>
</html>
